from typing import Optional

import click
import investor8_sdk
from pandas import DataFrame
from rich.console import Console

from i8_terminal.commands.earnings import earnings
from i8_terminal.common.cli import pass_command
from i8_terminal.common.formatting import get_formatter
from i8_terminal.common.layout import df2Table, format_df
from i8_terminal.common.utils import validate_ticker
from i8_terminal.types.ticker_param_type import TickerParamType


def get_upcoming_earnings_df(size: int) -> DataFrame:
    earnings = investor8_sdk.EarningsApi().get_upcoming_earnings(size=size)
    earnings = [d.to_dict() for d in earnings]
    df = DataFrame(earnings)
    return df


def get_upcoming_earnings_df_by_ticker(ticker: str) -> DataFrame:
    earnings = [investor8_sdk.EarningsApi().get_upcoming_earning(ticker).to_dict()]
    return DataFrame(earnings)


def format_upcoming_earnings_df(df: DataFrame, target: str) -> DataFrame:
    formatters = {
        "latest_price": get_formatter("number", target),
        "change": get_formatter("perc", target),
        "fyq": get_formatter("fyq", target),
        "eps_ws": get_formatter("number", target),
    }
    col_names = {
        "ticker": "Ticker",
        "name": "Name",
        "exchange": "Exchange",
        "sector": "Sector",
        "latest_price": "Price",
        "change": "Change",
        "actual_report_date": "Report Date",
        "fyq": "Period",
        "call_time": "Call Time",
        "eps_ws": "EPS Cons.",
    }
    return format_df(df, col_names, formatters)


def format_upcoming_earnings_df_by_ticker(df: DataFrame, target: str) -> DataFrame:
    formatters = {
        "actual_report_time": get_formatter("date", target),
        "fyq": get_formatter("fyq", target),
        "eps_ws": get_formatter("number", target),
        "revenue_ws": get_formatter("financials", target),
    }
    col_names = {
        "actual_report_time": "Date",
        "call_time": "Call Time",
        "fyq": "Period",
        "eps_ws": "EPS Cons.",
        "revenue_ws": "Revenue Cons.",
    }
    return format_df(df, col_names, formatters)


@earnings.command()
@click.option("--ticker", "-k", type=TickerParamType(), callback=validate_ticker, help="Company ticker.")
@pass_command
def upcoming(ticker: Optional[str]) -> None:
    """Lists upcoming company earnings."""
    console = Console()
    with console.status("Fetching data...", spinner="material"):
        df = get_upcoming_earnings_df_by_ticker(ticker) if ticker else get_upcoming_earnings_df(size=20)
    df_formatted = (
        format_upcoming_earnings_df_by_ticker(df, "console") if ticker else format_upcoming_earnings_df(df, "console")
    )
    table = df2Table(df_formatted)
    console.print(table)



from typing import List, Tuple

from i8_terminal.types.auto_complete_choice import AutoCompleteChoice


def get_indicators() -> List[Tuple[str, str]]:
    return [
        ("ma5", "5 Days Moving Average"),
        ("ma12", "12 Days Moving Average"),
        ("ma26", "26 Days Moving Average"),
        ("ma52", "26 Days Moving Average"),
        ("ema5", "5 Days Exponential Moving Average"),
        ("ema12", "12 Days Exponential Moving Average"),
        ("ema26", "26 Days Exponential Moving Average"),
        ("ema52", "52 Days Exponential Moving Average"),
        ("rsi", "Relative strength index (14 Days)"),
        ("rsi_7d", "Relative strength index (7 Days)"),
        ("rsi_1m", "Relative strength index (1 Month)"),
        ("rsi_3m", "Relative strength index (3 Months)"),
        ("alpha", "Alpha (1 Year)"),
        ("alpha_1w", "Alpha (1 Week)"),
        ("alpha_2w", "Alpha (2 Weeks)"),
        ("alpha_1m", "Alpha (1 Months)"),
        ("alpha_3m", "Alpha (3 Months)"),
        ("alpha_6m", "Alpha (6 Months)"),
        ("alpha_2y", "Alpha (2 Years)"),
        ("alpha_5y", "Alpha (5 Years)"),
        ("beta", "Beta (1 Year)"),
        ("beta_1w", "Beta (1 Week)"),
        ("beta_2w", "Beta (2 Weeks)"),
        ("beta_1m", "Beta (1 Months)"),
        ("beta_3m", "Beta (3 Months)"),
        ("beta_6m", "Beta (6 Months)"),
        ("beta_2y", "Beta (2 Years)"),
        ("beta_5y", "Beta (5 Years)"),
        ("volume", "Volume"),
    ]


class IndicatorParamType(AutoCompleteChoice):
    name = "indicator"

    def get_suggestions(self, keyword: str, pre_populate: bool = True) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.set_choices(get_indicators())

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]
        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "INDICATOR"



from typing import List, Tuple

from i8_terminal.common.metrics import get_all_metrics_df
from i8_terminal.types.auto_complete_choice import AutoCompleteChoice


def get_metrics() -> List[Tuple[str, str]]:
    df = get_all_metrics_df()[["metric_name", "name"]]
    return list(df.to_records(index=False))


class MetricParamType(AutoCompleteChoice):
    name = "metric"

    def get_suggestions(self, keyword: str, pre_populate: bool = True) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.set_choices(get_metrics())

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]

        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "METRIC"



from datetime import datetime
from typing import Any, Dict, List, Optional, cast

import click
import investor8_sdk
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from click.types import DateTime
from pandas import DataFrame
from plotly.subplots import make_subplots
from rich.console import Console

from i8_terminal.app.layout import get_date_range, get_plot_default_layout
from i8_terminal.app.plot_server import serve_plot
from i8_terminal.commands.metrics import metrics
from i8_terminal.common.cli import get_click_command_path, pass_command
from i8_terminal.common.metrics import get_daily_metrics_display_names
from i8_terminal.common.utils import PlotType, get_period_code
from i8_terminal.types.daily_metric_param_type import DailyMetricParamType
from i8_terminal.types.price_period_param_type import PricePeriodParamType
from i8_terminal.types.ticker_param_type import TickerParamType


def get_historical_daily_metrics_df(
    tickers: List[str], period: str, from_date: Optional[str], to_date: Optional[str]
) -> Optional[DataFrame]:
    period_code = get_period_code(period)
    hist_daily_metrics = []
    if from_date:
        if not to_date:
            to_date = datetime.now().strftime("%Y-%m-%d")
        for ticker in tickers:
            hist_daily_metrics.extend(
                investor8_sdk.MetricsApi().get_historical_daily_metrics(ticker, from_date=from_date, to_date=to_date)
            )
    else:
        for ticker in tickers:
            hist_daily_metrics.extend(
                investor8_sdk.MetricsApi().get_historical_daily_metrics(ticker, period=period_code)
            )
    if not hist_daily_metrics:
        return None
    df = DataFrame([d.to_dict() for d in hist_daily_metrics])
    df["date"] = pd.to_datetime(df["timestamp"], unit="s").dt.tz_localize("UTC")
    df = df.set_index("date")
    return df


def create_fig(
    df: DataFrame,
    period: str,
    cmd_context: Dict[str, Any],
    metrics: List[str],
    range_selector: bool = True,
) -> go.Figure:
    vertical_spacing = 0.02
    layout = dict(
        title=cmd_context["plot_title"],
        autosize=True,
        hovermode="closest",
        legend=dict(font=dict(size=11), orientation="v"),
        margin=dict(b=20, l=50, r=65),
    )
    rows_num = len(metrics)

    if rows_num == 2:
        row_width = [0.5, 0.5]
    else:
        row_width = [1]

    fig = make_subplots(
        rows=rows_num, cols=1, shared_xaxes=True, vertical_spacing=vertical_spacing, row_width=row_width
    )

    tickers = sorted(list(set(df["ticker"])))
    for m in metrics:
        for ticker, ticker_df in df.groupby("ticker"):
            fig.add_trace(
                go.Scatter(
                    x=ticker_df.index,
                    y=ticker_df[m],
                    name=ticker,
                    marker=dict(color=px.colors.qualitative.Plotly[tickers.index(ticker)]),
                    legendgroup=f"group{tickers.index(ticker)}",
                    showlegend=True if metrics.index(m) == 0 else False,
                ),
                row=metrics.index(m) + 1,
                col=1,
            )
    fig.update_traces(hovertemplate="%{y:.2f} %{x}")
    fig.update_xaxes(
        rangeslider_visible=False,
        spikemode="across",
        spikesnap="cursor",
    )

    if range_selector:
        fig.update_xaxes(rangeselector=get_date_range(get_period_code(period)))
    fig.update_layout(
        **layout,
        **get_plot_default_layout(),
        legend_title_text=None,
        xaxis_title=None,
        yaxis_title=None,
    )

    # Add yaxis titles
    for idx, r in enumerate(np.cumsum(row_width)[::-1]):
        fig["layout"][f"yaxis{idx+1}"]["title"] = get_daily_metrics_display_names(metrics)[idx]

    fig.update_annotations(
        dict(
            font_size=10,
            font_color="#525252",
        )
    )
    fig.update_yaxes(
        title_font_size=10,
    )

    return fig


@metrics.command()
@click.pass_context
@click.option("--tickers", "-k", type=TickerParamType(), required=True, help="Comma-separated list of tickers.")
@click.option(
    "--metrics",
    "-m",
    type=DailyMetricParamType(),
    default="pe_ratio_ttm",
    help="Comma-separated list of daily metrics.",
)
@click.option(
    "--period",
    "-p",
    type=PricePeriodParamType(),
    default="1Y",
    help="Historical price period.",
)
@click.option("--from_date", "-f", type=DateTime(), help="Histotical financials from date.")
@click.option("--to_date", "-t", type=DateTime(), help="Histotical financials to date.")
@pass_command
def plot(
    ctx: click.Context,
    tickers: str,
    metrics: str,
    period: str,
    from_date: Optional[datetime],
    to_date: Optional[datetime],
) -> None:
    """Compare and plot daily metrics of given companies. TICKERS is a comma seperated list of tickers.

    Examples:

    `i8 metrics plot --metrics pe_ratio_fy --period 5Y --tickers AMD,INTC,QCOM`
    """
    period = period.replace(" ", "").upper()
    metrics_list = metrics.replace(" ", "").split(",")
    if len(metrics_list) > 2:
        click.echo("You can enter up to 2 daily metrics.")
        return
    command_path_parsed_options_dict = {}
    if from_date:
        command_path_parsed_options_dict["--from_date"] = from_date.strftime("%Y-%m-%d")
    if to_date:
        command_path_parsed_options_dict["--to_date"] = to_date.strftime("%Y-%m-%d")
    command_path = get_click_command_path(ctx, command_path_parsed_options_dict)
    tickers_list = tickers.replace(" ", "").upper().split(",")
    if len(tickers_list) > 5:
        click.echo("You can enter up to 5 tickers.")
        return
    cmd_context = {
        "command_path": command_path,
        "tickers": tickers_list,
        "plot_type": PlotType.CHART.value,
    }

    console = Console()
    with console.status("Fetching data...", spinner="material") as status:
        df = get_historical_daily_metrics_df(tickers_list, period, cast(str, from_date), cast(str, to_date))
        cmd_context["plot_title"] = f"Historical daily {' and '.join(get_daily_metrics_display_names(metrics_list))}"
        status.update("Generating plot...")
        fig = create_fig(df, period, cmd_context, metrics_list, range_selector=False)

    serve_plot(fig, cmd_context)



import click
from rich.console import Console
from rich.table import Table

from i8_terminal.commands.metrics import metrics
from i8_terminal.common.cli import pass_command
from i8_terminal.common.metrics import get_metric_info
from i8_terminal.types.metric_param_type import MetricParamType


def get_metric_info_table(metric: str) -> Table:
    metric_display_name, metric_unit, metric_short_desc = get_metric_info(metric)
    metric_info_table = Table.grid()
    metric_info_table.add_column(justify="left", width=18)
    metric_info_table.add_column(justify="left")
    metric_info_table.add_row("Metric Name:", metric)
    metric_info_table.add_row("Display Name:", metric_display_name)
    metric_info_table.add_row("Unit:", metric_unit)
    metric_info_table.add_row("Description:", metric_short_desc)
    return metric_info_table


@metrics.command()
@click.option("--name", "-n", type=MetricParamType(), required=True, help="Metric name.")
@pass_command
def describe(name: str) -> None:
    """Describe the metric information.

    Examples:

    `i8 metrics describe --name basic_eps`
    """
    console = Console()
    metric_info = get_metric_info_table(name)
    console.print(metric_info)



import os
from typing import Dict, List, Optional, Tuple

import numpy as np
from investor8_sdk import MetricsApi
from pandas import DataFrame, read_csv

from i8_terminal.common.utils import is_cached_file_expired, similarity
from i8_terminal.config import APP_SETTINGS, SETTINGS_FOLDER


def get_indicators_list(indicator: Optional[str] = None) -> List[str]:
    indicators_dict = {
        "Momentum": ["MA5", "MA12", "MA26", "MA52", "EMA5", "EMA12", "EMA26", "EMA52"],
        "RSI": ["RSI_7D", "RSI_14D", "RSI_1M", "RSI_3M"],
        "Alpha": ["Alpha_1W", "Alpha_2W", "Alpha_1M", "Alpha_3M", "Alpha_6M", "Alpha_1Y", "Alpha_2Y", "Alpha_5Y"],
        "Beta": ["Beta_1W", "Beta_2W", "Beta_1M", "Beta_3M", "Beta_6M", "Beta_1Y", "Beta_2Y", "Beta_5Y"],
        "Volume": ["volume"],
    }
    if indicator:
        return indicators_dict.get(indicator, [""])
    else:
        return [item for sublist in indicators_dict.values() for item in sublist]


def find_similar_fin_metric(metric: str) -> Optional[str]:
    metrics_meta_data = get_all_metrics_df()[["metric_name", "name"]]
    best_match_similarity = 0.0
    best_match = ""
    for idx, m in metrics_meta_data.iterrows():
        sim = similarity(metric.upper(), m["metric_name"].upper())
        if sim > best_match_similarity:
            best_match_similarity = sim
            best_match = m["metric_name"]
    if best_match_similarity < APP_SETTINGS["metrics"]["similarity_threshold"]:
        return None
    return best_match.replace("_", "")


def find_similar_indicator(indicator: str) -> Optional[str]:
    default_indicators = {"RSI": "RSI_14D", "ALPHA": "Alpha_1Y", "BETA": "Beta_1Y", "VOLUME": "volume"}
    defualt_ind = default_indicators.get(indicator.upper())
    if defualt_ind:
        return defualt_ind
    indicators_list = get_indicators_list()
    best_match_similarity = 0.0
    best_match = ""
    for ind in indicators_list:
        sim = similarity(indicator.upper(), ind.upper())
        if sim > best_match_similarity:
            best_match_similarity = sim
            best_match = ind
    if best_match_similarity < APP_SETTINGS["metrics"]["similarity_threshold"]:
        return None

    return best_match


def get_all_metrics_df() -> DataFrame:
    metric_path = f"{SETTINGS_FOLDER}/metrics_metadata.csv"
    if os.path.exists(metric_path) and not is_cached_file_expired(metric_path):
        df = read_csv(metric_path)
    else:
        all_metrics = MetricsApi().get_list_financial_metrics_metadata()
        df = DataFrame([m.to_dict() for m in all_metrics])
        stat_code_mapping = {
            "income_statement": "inc",
            "cash_flow_statement": "cf",
            "balance_sheet_statement": "bs",
            "calculations": "calc",
        }
        df["tag_fullname"] = df.apply(lambda x: f"{x['tag']}_{stat_code_mapping[x['statement_code']]}", axis=1)
        df.to_csv(metric_path, index=False)

    return df


def get_metrics_display_names(metrics: List[str]) -> List[str]:
    all_metrics = get_all_metrics_df()[["metric_name", "name"]]
    all_metrics["metric_name"] = all_metrics["metric_name"].apply(lambda x: x.replace("_", ""))
    return list(set(all_metrics[all_metrics.metric_name.isin(metrics)]["name"]))


def get_all_daily_metrics() -> Dict[str, str]:
    return {
        "pe_ratio_fy": "P/E Ratio (FY)",
        "pe_ratio_ttm": "P/E Ratio (TTM)",
        "dividend_yield_fy": "Dividend Yield (FY)",
        "dividend_yield_ttm": "Dividend Yield (TTM)",
    }


def get_daily_metrics_display_names(metrics: List[str]) -> List[str]:
    all__daily_metrics = get_all_daily_metrics()
    return list(all__daily_metrics[metric] for metric in metrics)


def get_metric_info(name: str) -> Tuple[str, str, str]:
    all_metrics = get_all_metrics_df()
    metric = all_metrics.loc[all_metrics.metric_name == name].replace(np.nan, "", regex=True).to_dict("records")[0]
    return (
        metric["name"],
        metric["unit"],
        metric["short_description"]
        if "short_description" in metric.keys() and metric["short_description"]
        else "No Description",
    )



import enum
import os
from difflib import SequenceMatcher
from typing import Any, Dict, Optional

import arrow
import click
import pandas as pd
from rich.console import Console

from i8_terminal.config import APP_SETTINGS


class PlotType(enum.Enum):
    CHART = "chart"
    TABLE = "table"


def to_snake_case(value: str) -> str:
    return "_".join(value.lower().split())


def get_period_code(period: str) -> int:
    return {"1D": 1, "5D": 2, "1M": 3, "3M": 4, "6M": 5, "1Y": 6, "3Y": 7, "5Y": 8}.get(period, 3)


def get_period_days(period: str) -> int:
    return {"1D": 1, "5D": 5, "1M": 30, "3M": 90, "6M": 180, "1Y": 365, "3Y": 1095, "5Y": 1825}.get(period, 365)


def similarity(a: str, b: str) -> float:
    return SequenceMatcher(None, a, b).ratio()


def export_data(
    export_df: pd.DataFrame,
    export_path: str,
    column_width: Optional[int],
    column_format: Dict[str, Any],
    index: bool = False,
) -> None:
    console = Console()
    extension = export_path.split(".")[-1]
    if extension == "csv":
        export_df.to_csv(export_path, index=index)
        console.print(f"Data is saved on: {export_path}")
    elif extension == "xlsx":
        writer = pd.ExcelWriter(export_path, engine="xlsxwriter")
        export_df.to_excel(writer, sheet_name="Sheet1", startrow=1, header=False, index=index)
        workbook = writer.book
        worksheet = writer.sheets["Sheet1"]
        header_format = workbook.add_format(APP_SETTINGS["styles"]["xlsx"]["default"]["header"])
        metric_format = workbook.add_format(APP_SETTINGS["styles"]["xlsx"]["default"]["metric"])
        column_format = workbook.add_format(column_format)
        headers = export_df.columns.tolist()
        if index:
            headers.insert(0, (export_df.index.name, ""))
        for col_num, value in enumerate(headers):
            if type(value) is tuple:
                worksheet.merge_range(0, col_num, len(value) - 1, col_num, " ".join(reversed(value)), header_format)
            else:
                worksheet.write(0, col_num, value, header_format)
        worksheet.set_column(0, 0, 20, metric_format)
        worksheet.set_column(
            1, len(export_df.columns) - 1 if not index else len(export_df.columns), column_width, column_format
        )
        writer.save()
        console.print(f"Data is saved on: {export_path}")
    else:
        console.print("export_path is not valid")


def validate_ticker(ctx: click.Context, param: str, value: str) -> Optional[str]:
    if value and len(value.replace(" ", "").split(",")) > 1:
        click.echo(click.style(f"`{value}` is not a valid ticker name.", fg="yellow"))
        ctx.exit()
    return value


def is_cached_file_expired(file_path: str) -> bool:
    mtime = arrow.get(os.path.getmtime(file_path))
    return bool(mtime < arrow.utcnow().shift(hours=-APP_SETTINGS.get("cache", {}).get("age", 48)))
